/***************************************************************************
 LIBNOTEMIDI - API for accessing the NOTEMIDI Linux serial port MIDI module

	(C) 2000 by Michael Minn - nycminn@earthlink.net

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <sys/io.h>
#include <sys/time.h>
#include <sys/stat.h> // for fstat()
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include "libnotemidi.h"

libnotemidi::libnotemidi(char *notemidi_port)
{
	if ((file_descriptor = open(notemidi_port, O_RDWR | O_NOCTTY)) < 0)
		perror("Failure opening NOTEMIDI port");
}

libnotemidi::~libnotemidi()
{
	if (file_descriptor >= 0)
		close(file_descriptor);
}

int libnotemidi::read_midi_file(char *midi_file_name, libnotemidi_file &midi_file)
{
	int status = 0;
	unsigned char *buffer = NULL;
	FILE *disk_file = fopen(midi_file_name, "r");
	struct stat file_info;
	
	if (disk_file == NULL)
		perror(midi_file_name);
		
	else if (stat(midi_file_name, &file_info))
		perror(midi_file_name);
		
	else if ((buffer = new unsigned char[file_info.st_size]) == NULL)
		perror("memory allocation failure");
		
	else if (fread(buffer, 1, file_info.st_size, disk_file) < file_info.st_size)
		perror("failure reading midi file");
	
	else if (midi_file.load(buffer, file_info.st_size) == NULL)
		printf("failure interpreting midi file");
	else
		status = file_info.st_size;
	
	if (disk_file)
		fclose(disk_file);
		
	if (buffer)
		delete buffer;
		
	return status;		
}

int libnotemidi::play_midi_file(char *midi_file_name)
{
	int status = 0;
	libnotemidi_file midi_file;
	struct timeval start_time;
	struct timezone zone;
	
	if (!is_ok())
		printf("MIDI port not opened\n");
		
	else if (read_midi_file(midi_file_name, midi_file) <= 0)
		;
		
	else if (gettimeofday(&start_time, &zone))
		perror("failure getting start time");
		
	else for (libnotemidi_event *event = midi_file.events->find_first_event(); event && (status >= 0); )
	{
		struct timeval current_time;
		gettimeofday(&current_time, &zone);
		// need multiple computations to kludge problem with unsigned int
		int current_seconds = current_time.tv_sec - start_time.tv_sec;
		int current_nanoseconds = 1000000 + current_time.tv_usec - start_time.tv_usec - 1000000;
		int current_milliseconds = (current_seconds * 1000) + (current_nanoseconds / 1000);
		
		while (event && (status >= 0) && (event->absolute_milliseconds <= current_milliseconds))
		{
			if ((event->data[0] == LIBNOTEMIDI_META_EVENT) &&
			    ((event->data[1] == LIBNOTEMIDI_META_TEXT) ||
			     (event->data[1] == LIBNOTEMIDI_META_COPYRIGHT) ||
			     (event->data[1] == LIBNOTEMIDI_META_SEQUENCE_NAME) ||
			     (event->data[1] == LIBNOTEMIDI_META_LYRIC)))
			     printf("%*s\n", (int) event->data[2], (char*) &event->data[3]);

			else if (event->data[0] <= LIBNOTEMIDI_SYSTEM_EXCLUSIVE)
				status = write(file_descriptor, event->data, event->find_length());
			event = event->next;			
		}
		
		if (event)
		{
			gettimeofday(&current_time, &zone);
			current_seconds = current_time.tv_sec - start_time.tv_sec;
			current_nanoseconds = 1000000 + current_time.tv_usec - start_time.tv_usec - 1000000;
			current_milliseconds = (current_seconds * 1000) + (current_nanoseconds / 1000);
			int sleep_milliseconds = event->absolute_milliseconds - current_milliseconds;
			if (sleep_milliseconds > 0)
				usleep(sleep_milliseconds * 1000);
		}
	}
	
	return status;
}

static int interrupt = 0;
static void interrupt_handler(int signal_number) { ++interrupt; }
	
int libnotemidi::record_midi_file(char *midi_file_name)
{
	int status = 1;
	int length = 0;
	int beats_per_minute = 120; // default tempo
	unsigned char buffer[3];
	unsigned char running_status = 0;
	libnotemidi_file midi_file;
	FILE *output_file = fopen(midi_file_name, "w");
	struct timeval previous_time;
	struct timezone zone;
	
	signal(SIGINT, &interrupt_handler);
	interrupt = 0;
	
	if (output_file == NULL)
		perror(midi_file_name);
		
	else if (!is_ok())
		printf("MIDI port not opened\n");
		
	else if (gettimeofday(&previous_time, &zone))
		perror("failure getting start time");
		
	else while (interrupt == 0)
	{
		length = 0;
		status = 0;
		buffer[0] = 0;
		
		while ((interrupt == 0) && ((status <= 0) || ((int)buffer[0] >= LIBNOTEMIDI_SYSTEM_EXCLUSIVE)))
		       status = read(file_descriptor, &buffer[0], 1);
		
		if (status && (buffer[0] < LIBNOTEMIDI_NOTE_OFF))
		{
			buffer[1] = buffer[0];
			buffer[0] = running_status;
			if (status && (buffer[0] < LIBNOTEMIDI_PROGRAM_CHANGE))
			{
			       status = read(file_descriptor, &buffer[2], 1);
			       length = 3;			
			}
			else
				length = 2;
		}
		
		else if (status && (buffer[0] >= LIBNOTEMIDI_PROGRAM_CHANGE) && (buffer[0] < LIBNOTEMIDI_SYSTEM_EXCLUSIVE))
		{
		       running_status = buffer[0];
		       status = read(file_descriptor, &buffer[1], 1);
		       length = 2;
		}
		
		else if (status)
		{
		       running_status = buffer[0];
		       status = read(file_descriptor, &buffer[1], 1);
		       status = read(file_descriptor, &buffer[2], 1);
		       length = 3;
		}
		
		if (length)
		{
			libnotemidi_event *event = new libnotemidi_event;
			event->data = new unsigned char[length];
			memcpy(event->data, buffer, length);
			
			struct timeval current_time;
			gettimeofday(&current_time, &zone);
			// need multiple computations to kludge problem with unsigned int
			int delta_seconds = current_time.tv_sec - previous_time.tv_sec;
			int delta_nanoseconds = 1000000 + current_time.tv_usec - previous_time.tv_usec - 1000000;
			int delta_milliseconds = (delta_seconds * 1000) + (delta_nanoseconds / 1000);
			event->delta_time = delta_milliseconds * midi_file.find_division() * beats_per_minute / 60 / 1000;
			previous_time = current_time;
	
			/* printf("%d: %02x%02x%02x\n", event->delta_time, (int)event->data[0], (int)event->data[1], (int)event->data[2]); */
	
			midi_file.events = midi_file.events->add_event(event);
			status = write(file_descriptor, buffer, length);
		}			
	}

	unsigned char file_buffer[256000];
	int file_length = midi_file.save(file_buffer);

	if (fwrite(file_buffer, 1, file_length, output_file) < file_length)
		perror("Failure writing MIDI file\n");
	else
		printf("MIDI file %s written\n", midi_file_name);

	if (output_file)
		fclose(output_file);
	
	return 1;
}


// NOTE: the following libnotemidi_event & file structures have
// intentionally been made system independent (i.e. no system calls)
// to enhance portability

/***************** General MIDI Names *****************/

static char* general_midi_programs[129] = {
	"0 acoustic grand piano",
	"1 bright acoustic piano",
	"2 electric grand piano",
	"3 honky-tonk piano",
	"4 rhodes piano",
	"5 chorused piano",
	"6 harpsichord",
	"7 clavinet",

	"8 celesta",
	"9 glockenspiel",
	"10 music box",
	"11 vibraphone",
	"12 marimba",
	"13 xylophone",
	"14 tubular bells",
	"15 dulcimer",

	"16 hammond organ",
	"17 percussive organ",
	"18 rock organ",
	"19 church organ",
	"20 reed organ",
	"21 accordian",
	"22 harmonica",
	"23 tango accordian",

	"24 nylon acoustic guitar",
	"25 steel acoustic guitar",
	"26 electric guitar (clean)",
	"27 electric guitar (jazz)",
	"28 electric guitar (muted)",
	"29 overdriven guitar",
	"30 distortion guitar",
	"31 guitar harmonics",

	"32 acoustic bass",
	"33 electric bass (fingered)",
	"34 electric bass (picked)",
	"35 fretless bass",
	"36 slap bass 1",
	"37 slap bass 2",
	"38 synth bass 1",
	"39 synth bass 2",

	"40 violin",
	"41 viola",
	"42 cello",
	"43 contrabass",
	"44 tremelo strings",
	"45 pizzicato strings",
	"46 orchestral harp",
	"47 timpani",

	"48 string ensemble",
	"49 slow strings",
	"50 synth strings 1",
	"51 synth strings 2",
	"52 choir aahs",
	"53 voice oohs",
	"54 synth voice",
	"55 orchestra hit",

	"56 trumpet",
	"57 trombone",

	"58 tuba",
	"59 muted trumpet",
	"60 french horn",
	"61 brass section",
	"62 synth brass 1",
	"63 synth brass 2",

	"64 soprano sax",
	"65 alto sax",
	"66 tenor sax",
	"67 baritone sax",
	"68 oboe",
	"69 english horn",
	"70 bassoon",
	"71 clarinet",

	"72 piccolo",
	"73 flute",
	"74 recorder",
	"75 pan flute",
	"76 blown bottle",
	"77 shakahachi",
	"78 whistle",
	"79 ocarina",

	"80 square lead 1",
	"81 saw lead 2",
	"82 calliope lead 3",
	"83 chiffer lead 4",
	"84 charang lead 5",
	"85 syn vox lead 6",
	"86 5th sawtooth lead 7",
	"87 bass & lead 8",

	"88 fantasia pad 1",
	"89 warm pad 2",
	"90 polysynth pad 3",
	"91 space voice pad 4",
	"92 bowed glass pad 5",
	"93 metal pad 6",
	"94 halo pad 7",
	"95 sweep pad 8",

	"96 ice rain fx 1",
	"97 soundtrack fx 2",
	"98 crystal fx 3",
	"99 atmosphere fx 4",
	"100 brightness fx 5",
	"101 goblin fx 6",
	"102 echo drops fx 7",
	"103 star theme fx 8",

	"104 sitar",
	"105 banjo",
	"106 shamisen",


	"107 koto",
	"108 kalimba",

	"109 bagpipe",
	"110 fiddle",
	"111 shanai",

	"112 tinkle bell",
	"113 agogo",
	"114 steel drums",
	"115 woodblock",
	"116 taiko drum",
	"117 melodic tom",
	"118 synth drum",
	"119 reverse cymbal",

	"120 guitar-fret noise",
	"121 breath noise",
	"122 seashore",
	"123 bird tweet",
	"124 telephone ring",
	"125 helicopter",
	"126 applause",
	"127 gunshot",
	NULL };

static char* general_midi_status_bytes[129] = {
	"Chan 1 Note Off",
	"Chan 2 Note Off",
	"Chan 3 Note Off",
	"Chan 4 Note Off",
	"Chan 5 Note Off",
	"Chan 6 Note Off",
	"Chan 7 Note Off",
	"Chan 8 Note Off",
	"Chan 9 Note Off",
	"Chan 10 Note Off",
	"Chan 11 Note Off",
	"Chan 12 Note Off",
	"Chan 13 Note Off",
	"Chan 14 Note Off",
	"Chan 15 Note Off",
	"Chan 16 Note Off",
	"Chan 1 Note On ",
	"Chan 2 Note On",
	"Chan 3 Note On",
	"Chan 4 Note On",
	"Chan 5 Note On",
	"Chan 6 Note On",
	"Chan 7 Note On",
	"Chan 8 Note On",
	"Chan 9 Note On",
	"Chan 10 Note On",
	"Chan 11 Note On",
	"Chan 12 Note On",
	"Chan 13 Note On",
	"Chan 14 Note On",
	"Chan 15 Note On",
	"Chan 16 Note On",
	"Chan 1 Polyphonic Aftertouch",
	"Chan 2 Polyphonic Aftertouch",
	"Chan 3 Polyphonic Aftertouch",
	"Chan 4 Polyphonic Aftertouch",
	"Chan 5 Polyphonic Aftertouch",
	"Chan 6 Polyphonic Aftertouch",
	"Chan 7 Polyphonic Aftertouch",
	"Chan 8 Polyphonic Aftertouch",
	"Chan 9 Polyphonic Aftertouch",
	"Chan 10 Polyphonic Aftertouch",
	"Chan 11 Polyphonic Aftertouch",
	"Chan 12 Polyphonic Aftertouch",
	"Chan 13 Polyphonic Aftertouch",
	"Chan 14 Polyphonic Aftertouch",
	"Chan 15 Polyphonic Aftertouch",
	"Chan 16 Polyphonic Aftertouch",
	"Chan 1 Controller",
	"Chan 2 Controller",
	"Chan 3 Controller",
	"Chan 4 Controller",
	"Chan 5 Controller",
	"Chan 6 Controller",
	"Chan 7 Controller",
	"Chan 8 Controller",
	"Chan 9 Controller",
	"Chan 10 Controller",
	"Chan 11 Controller",
	"Chan 12 Controller",
	"Chan 13 Controller",
	"Chan 14 Controller",
	"Chan 15 Controller",
	"Chan 16 Controller",
	"Chan 1 Program Change",
	"Chan 2 Program Change",
	"Chan 3 Program Change",
	"Chan 4 Program Change",
	"Chan 5 Program Change",
	"Chan 6 Program Change",
	"Chan 7 Program Change",
	"Chan 8 Program Change",
	"Chan 9 Program Change",
	"Chan 10 Program Change",
	"Chan 11 Program Change",
	"Chan 12 Program Change",
	"Chan 13 Program Change",
	"Chan 14 Program Change",
	"Chan 15 Program Change",
	"Chan 16 Program Change",
	"Chan 1 Channel Aftertouch",
	"Chan 2 Channel Aftertouch",
	"Chan 3 Channel Aftertouch",
	"Chan 4 Channel Aftertouch",
	"Chan 5 Channel Aftertouch",
	"Chan 6 Channel Aftertouch",
	"Chan 7 Channel Aftertouch",
	"Chan 8 Channel Aftertouch",
	"Chan 9 Channel Aftertouch",
	"Chan 10 Channel Aftertouch",
	"Chan 11 Channel Aftertouch",
	"Chan 12 Channel Aftertouch",
	"Chan 13 Channel Aftertouch",
	"Chan 14 Channel Aftertouch",
	"Chan 15 Channel Aftertouch",
	"Chan 16 Channel Aftertouch",
	"Chan 1 Pitch Wheel",
	"Chan 2 Pitch Wheel",
	"Chan 3 Pitch Wheel",
	"Chan 4 Channel Aftertouch",
	"Chan 5 Channel Aftertouch",
	"Chan 6 Channel Aftertouch",
	"Chan 7 Channel Aftertouch",
	"Chan 8 Channel Aftertouch",
	"Chan 9 Channel Aftertouch",
	"Chan 10 Channel Aftertouch",
	"Chan 11 Channel Aftertouch",
	"Chan 12 Channel Aftertouch",
	"Chan 13 Channel Aftertouch",
	"Chan 14 Channel Aftertouch",
	"Chan 15 Channel Aftertouch",
	"Chan 16 Channel Aftertouch",
	"System Exclusive (F0)",
	"MIDI Time Code 1/4 Frame (F1)",
	"Song Position Pointer (F2)",
	"Song # Select (F3)",
	"Undefined (F4)",
	"Undefined (F5)",
	"Tune request (F6)",
	"End of SysEx (EOX) (F7)",
	"Timing clock (F8)",
	"Undefined (F9)",
	"Start (FA)",
	"Continue (FB)",
	"Stop (FC)",
	"Undefined (FD)",
	"Active Sensing (FE)",
	"Meta (FF)",
	NULL
};

int libnotemidi_event::set_absolute_times(int ticks_per_quarter_note)
{
	int current_milliseconds = 0;
	int current_ticks_per_second = ticks_per_quarter_note * 2; // default tempo 120 beats/second

	for (libnotemidi_event *event = find_first_event(); event != NULL; event = event->next)
	{
		int ticks_since_previous = 0;
		if (event->previous)
			ticks_since_previous = event->absolute_time - event->previous->absolute_time;
		current_milliseconds += (ticks_since_previous * 1000 / current_ticks_per_second);
		event->absolute_milliseconds = current_milliseconds;

//char string[256];
//printf("%03d %s\n", event->absolute_time, event->find_name(string));
//usleep(100000);

		if (event->is_tempo())
			current_ticks_per_second = ticks_per_quarter_note * event->find_tempo_beats_per_minute() / 60;
	}
	
	return current_milliseconds;
}

int libnotemidi_event::find_tempo_beats_per_minute()
{
	if (this && data && (data[0] == LIBNOTEMIDI_META_EVENT) && (data[1] == LIBNOTEMIDI_META_TEMPO))
	{
		int microseconds_per_quarter_note = long_from_bytes(3, &data[3]);
		if (microseconds_per_quarter_note > 0)
			return 60000000 / microseconds_per_quarter_note;
	}
	
	return 0;
}

int libnotemidi_event::find_length()
{
	switch (data[0] & 0xF0)
	{
		case LIBNOTEMIDI_NOTE_OFF:
		case LIBNOTEMIDI_NOTE_ON:
		case LIBNOTEMIDI_POLYPHONIC_KEY_PRESSURE:
		case LIBNOTEMIDI_CONTROL_CHANGE:
		case LIBNOTEMIDI_PITCH_BEND:
			return 3;

		case LIBNOTEMIDI_PROGRAM_CHANGE:
		case LIBNOTEMIDI_AFTERTOUCH:
			return 2;

		default:
			switch (data[0])
			{
				case LIBNOTEMIDI_SONG_POSITION:
					return 3;

				case LIBNOTEMIDI_TIME_CODE_QUARTER_FRAME:
				case LIBNOTEMIDI_SONG_SELECT:
					return 2;

				case LIBNOTEMIDI_TUNE_REQUEST:
				case LIBNOTEMIDI_END_OF_EXCLUSIVE:
				case LIBNOTEMIDI_TIMING_CLOCK:
				case LIBNOTEMIDI_START:
				case LIBNOTEMIDI_CONTINUE:
				case LIBNOTEMIDI_STOP:
				case LIBNOTEMIDI_ACTIVE_SENSING:
					return 1;

				case LIBNOTEMIDI_META_EVENT:
					return data[2] + 3;

				case LIBNOTEMIDI_SYSTEM_EXCLUSIVE:
					int scan = 0;

					while ((scan < 255) && (data[scan] != LIBNOTEMIDI_END_OF_EXCLUSIVE))
						++scan;

					return scan + 1;
			}
	}

	return 0;
}

void libnotemidi_event::delete_list()
{
	if (this == NULL)
		return;

	libnotemidi_event *next_event = NULL;
	for (libnotemidi_event *scan = find_first_event(); scan != NULL; scan = next_event)
	{
		next_event = scan->next;
		delete scan;
	}
}

char *libnotemidi_event::find_name(char *string)
{
	//sscanf(string, "%03d:%03d:%03d(%06d) ", measure, time, offset, delta_time);
	//int string_offset = strlen(string);
	//int channel_offset = (data[0] & 0x0F) + 1 + string_offset;
	//strcat(string, "                  ");

	sprintf(string, "%d:%03d ", absolute_milliseconds / 1000, absolute_milliseconds % 1000);

	switch (data[0] & 0xF0)
	{
		case LIBNOTEMIDI_NOTE_OFF: sprintf(&string[strlen(string)], "Note Off ch %d, pitch %d, vel %d", (int)data[0] & 0x0f, (int)data[1], (int)data[2]); break;
		case LIBNOTEMIDI_NOTE_ON: sprintf(&string[strlen(string)],  "Note On  ch %d, pitch %d, vel %d", (int)data[0] & 0x0f, (int)data[1], (int)data[2]); break;
		case LIBNOTEMIDI_POLYPHONIC_KEY_PRESSURE: sprintf(&string[strlen(string)], "Aftertouch ch %d, %d, %d", (int)data[0] & 0x0f, (int)data[1], (int)data[2]); break;
		case LIBNOTEMIDI_PITCH_BEND: sprintf(&string[strlen(string)], "Pitch bend ch %d, amount %d", (int)data[0] & 0x0f, (int) long_from_14bit(&data[1])); break;
		case LIBNOTEMIDI_PROGRAM_CHANGE: sprintf(&string[strlen(string)], "Program ch %d, %s", (int)data[0] & 0x0f, general_midi_programs[data[1] & 0x7f]); break;
		case LIBNOTEMIDI_CONTROL_CHANGE: sprintf(&string[strlen(string)], "Controller %d = %d, ch %d", (int)data[1], (int)data[2], (int)data[0] & 0x0f); break;
		case LIBNOTEMIDI_AFTERTOUCH: sprintf(&string[strlen(string)], "Channel %d aftertouch %d", (int)data[0] & 0x0f, (int)data[0]); break;
	}

	switch (data[0])
	{
		case LIBNOTEMIDI_TIME_CODE_QUARTER_FRAME: sprintf(&string[strlen(string)],  "MIDI time code 1/4 frame"); break;
		case LIBNOTEMIDI_SONG_POSITION: sprintf(&string[strlen(string)], "Song position %d", (int) long_from_14bit(&data[1])); break;
		case LIBNOTEMIDI_SONG_SELECT: sprintf(&string[strlen(string)], "Song select number %d", (int)data[1]); break;

		case LIBNOTEMIDI_TUNE_REQUEST: sprintf(&string[strlen(string)], "Tune request"); break;
		case LIBNOTEMIDI_END_OF_EXCLUSIVE: sprintf(&string[strlen(string)], "End of system exclusive"); break;
		case LIBNOTEMIDI_TIMING_CLOCK: sprintf(&string[strlen(string)], "Timing clock"); break;
		case LIBNOTEMIDI_START: sprintf(&string[strlen(string)], "Start"); break;
		case LIBNOTEMIDI_CONTINUE: sprintf(&string[strlen(string)], "Continue"); break;
		case LIBNOTEMIDI_STOP: sprintf(&string[strlen(string)], "Stop"); break;
		case LIBNOTEMIDI_ACTIVE_SENSING: sprintf(&string[strlen(string)], "Active sensing"); break;
		case LIBNOTEMIDI_META_EVENT: find_meta_event_name(&string[strlen(string)]); break;
	}

	if (data[0] == LIBNOTEMIDI_SYSTEM_EXCLUSIVE)
	{
		int index;
		unsigned char data_value = 0;
		strcpy(&string[strlen(string)], "Sys Ex ");

		for (int scan = 0; (strlen(string) < 128) && ((scan == 0) || (data[scan - 1] != LIBNOTEMIDI_END_OF_EXCLUSIVE)); ++scan)
			sprintf(&string[strlen(string)], " %02X", (int)data[scan]);
	}

	return string;
}

char *libnotemidi_event::find_meta_event_name(char *string)
{
	if ((data == NULL) || (data[0] != LIBNOTEMIDI_META_EVENT))
		return NULL;

	switch (data[1])
	{
		case LIBNOTEMIDI_META_SEQUENCE: sprintf(string, "Sequence #%d", (int) long_from_bytes(2, &data[3])); break;
		case LIBNOTEMIDI_META_TEXT: strcpy(string, "Text: "); string[strlen(string) + data[2]] = (char)NULL; strncat(string, (char*)&data[3], (int)data[2]); break;
		case LIBNOTEMIDI_META_COPYRIGHT: strcpy(string, "Copyright: "); string[strlen(string) + data[2]] = (char)NULL; strncat(string, (char*)&data[3], (int)data[2]);  break;
		case LIBNOTEMIDI_META_SEQUENCE_NAME: strcpy(string, "Sequence/track name: "); string[strlen(string) + data[2]] = (char)NULL; strncat(string, (char*)&data[3], (int)data[2]); break;
		case LIBNOTEMIDI_META_INSTRUMENT_NAME: strcpy(string, "Instrument name: "); string[strlen(string) + data[2]] = (char)NULL; strncat(string, (char*)&data[3], (int)data[2]); break;
		case LIBNOTEMIDI_META_LYRIC: strcpy(string, "Lyric: "); string[strlen(string) + data[2]] = (char)NULL; strncat(string, (char*)&data[3], (int)data[2]); break;
		case LIBNOTEMIDI_META_MARKER: strcpy(string, "Marker: "); string[strlen(string) + data[2]] = (char)NULL; strncat(string, (char*)&data[3], (int)data[2]); break;
		case LIBNOTEMIDI_META_CUE_POINT: strcpy(string, "Cue point: "); string[strlen(string) + data[2]] = (char)NULL; strncat(string, (char*)&data[3], (int)data[2]); break;
		case LIBNOTEMIDI_META_CHANNEL_PREFIX: sprintf(string, "MIDI channel prefix %d", (int)data[3]); break;
		case LIBNOTEMIDI_META_END_OF_TRACK: sprintf(string, "End of track"); break;
		case LIBNOTEMIDI_META_TEMPO: sprintf(string, "Tempo: %d quarter notes/second", 60000000 / (int) long_from_bytes(3, &data[3])); break;
		case LIBNOTEMIDI_META_TIME_SIGNATURE: sprintf(string, "Time signature: %d/%d metronome %d, 32nds/quarter %d", (int)data[3], (1 << (int)data[4]), (int)data[5], (int)data[6]); break;
		case LIBNOTEMIDI_META_KEY_SIGNATURE: sprintf(string, "Key %d, minor %d", (int)data[3], (int)data[4]); break;
		case LIBNOTEMIDI_META_SEQUENCER_SPECIFIC: sprintf(string, "Sequencer Specific Meta: %*s", (int) data[2], &data[3]); break;
		default: sprintf(string, "Meta event %d", (int) data[2]); break;
	}

	return string;
}

libnotemidi_event *libnotemidi_event::find_first_event()
{
	if (this == NULL)
		return NULL;


	libnotemidi_event *scan = this;
	while (scan->previous != NULL)
		scan = scan->previous;

	return scan;
}

libnotemidi_event *libnotemidi_event::add_event(libnotemidi_event *event)
{
	libnotemidi_event *scan = this;

	if (this == NULL)
	{
		event->next = NULL;
		event->previous = NULL;
	}
	else if (event->absolute_time >= this->absolute_time)
	{
		while (scan->next && (event->absolute_time >= scan->next->absolute_time))
			scan = scan->next;
		event->next = scan->next;
		event->previous = scan;
		scan->next = event;
		if (event->next != NULL)
			event->next->previous = event;
	}
	else
	{
		while (scan->previous && (event->absolute_time < scan->previous->absolute_time))
			scan = scan->previous;
		event->previous = scan->previous;
		event->next = scan;
		scan->previous = event;
		if (event->previous != NULL)
			event->previous->next = event;
	}

	return event;
}

unsigned char *libnotemidi_event::write_delta_time(unsigned char *buffer)
{
	int bytes = find_number_of_delta_time_bytes();

	for (int scan = bytes; scan > 0; --scan)
	{
		int byte = (delta_time >> (7 * (scan - 1))) & 0x7f;

		if (scan > 1)
			buffer[0] = ((char)byte & 0xff) | 0x80;
		else
			buffer[0] = ((char)byte & 0xff);

		++buffer;
	}

	return buffer;
}


unsigned char *libnotemidi_event::write_event(unsigned char *buffer)
{
	buffer = write_delta_time(buffer);

	if (data[0] == LIBNOTEMIDI_SYSTEM_EXCLUSIVE)
	{
		// System exclusive events have a length byte in MIDI files after 0xf0
		// See "The Musicians Guide to MIDI" pp 362
		buffer[0] = data[0];
		buffer[1] = find_length() - 1;

		memcpy(&buffer[2], &data[1], (int) buffer[1]);
		buffer += (((int) buffer[1]) + 2);
	}
	else
	{
		memcpy(buffer, data, find_length());
		buffer += find_length();
	}

	return buffer;
}

unsigned char *libnotemidi_event::save_events(unsigned char *buffer)
{
	for (libnotemidi_event *scan = find_first_event(); scan != NULL; scan = scan->next)
		buffer = scan->write_event(buffer);

	return buffer;
}

unsigned char *libnotemidi_event::read_delta_time(unsigned char *buffer)
{
	delta_time = 0;

	do

	{
		delta_time <<= 7;
		delta_time += ((*buffer) & 0x7F);
		++buffer;
	}
	while (((*(buffer - 1)) & 0x80) != 0);

	return buffer;
}

unsigned char *libnotemidi_event::read_event(unsigned char *buffer, unsigned char *buffer_end, unsigned char &current_status_byte)
{
	if ((buffer[0] & 0x80) == (char)NULL) /* see "Running status" pp156 */
	{
		--buffer;
		buffer[0] = current_status_byte;
	}
	else
		current_status_byte = buffer[0];

	unsigned char event_length = 0;

	if (buffer[0] == LIBNOTEMIDI_META_EVENT)
		event_length = buffer[2] + 3;

	else if (buffer[0] == LIBNOTEMIDI_SYSTEM_EXCLUSIVE)
	{
		unsigned char total = 1;
		while ((buffer[total] != LIBNOTEMIDI_END_OF_EXCLUSIVE) && (&buffer[total] < buffer_end))
			++total;
		event_length = total + 1;
	}

	else
	{
		data = buffer;
		event_length = (unsigned char)(find_length() & 0x0f);
		data = NULL;
	}

	if ((buffer + event_length) > buffer_end)
		return NULL; // file format error
	else if (buffer[0] == LIBNOTEMIDI_SYSTEM_EXCLUSIVE)
	{
		// midi files have a length byte after the initial 0xf0
		data = new unsigned char[event_length];
		data[0] = LIBNOTEMIDI_SYSTEM_EXCLUSIVE;
		memcpy(&data[1], &buffer[2], event_length - 2);
		return buffer + event_length;
	}
	else
	{
		data = new unsigned char[event_length];
		memcpy(data, buffer, event_length);
		return buffer + event_length;
	}
}

libnotemidi_event *libnotemidi_event::load_events(unsigned char *(&buffer), int buffer_length, int &track_start_time)
{
	//unsigned char *buffer_start = buffer;
	unsigned char *buffer_end = buffer + buffer_length;
	libnotemidi_event *events = this;
	unsigned char current_status_byte = 0;
	int current_time = track_start_time;

	while ((buffer != NULL) && (buffer < buffer_end))
	{
		libnotemidi_event *event = new libnotemidi_event;
		buffer = event->read_delta_time(buffer);
		current_time += event->delta_time;
		event->absolute_time = current_time;

		buffer = event->read_event(buffer, buffer_end, current_status_byte);
		events = events->add_event(event);
	}

	track_start_time = current_time;
	
	return events;
}

int libnotemidi_event::long_from_bytes(int byte_count, unsigned char *bytes)
{
	int value = 0;
	for (int scan = 0; scan < byte_count; ++scan)
    {
		value <<= 8;
		value += bytes[scan];
	}
	return value;

}

void libnotemidi_event::long_to_bytes(int byte_count, unsigned char *bytes, int value)
{
	for (int scan = byte_count - 1; scan >= 0; --scan)
    {
		bytes[scan] = (unsigned char) (value & 0xff);
		value >>= 8;
	}
}

/***************** LIBNOTEMIDI_FILE ***************/

int libnotemidi_file::save(unsigned char *buffer)
{
	if ((this == NULL) || (buffer == NULL))
		return 0;

	unsigned char *buffer_head = buffer;
	/* events->set_midi_times(file_header.division); */

	buffer = write_file_header(buffer);

	buffer = write_track_header(buffer, 0);

	buffer = events->save_events(buffer);

	buffer = write_end_of_track(buffer);

	int file_length = buffer - buffer_head;
	int track_length = file_length - sizeof(libnotemidi_character_file_header) - sizeof(libnotemidi_character_track_header);
	unsigned char *track_header = &buffer_head[sizeof(libnotemidi_character_file_header)];

	write_track_header(track_header, track_length);
	return file_length;
}

unsigned char *libnotemidi_file::load(unsigned char *buffer, int buffer_length)
{
	int current_absolute_time = 0;
	unsigned char *buffer_end = buffer + buffer_length;

	buffer = read_file_header(buffer);

	for (unsigned int track = 0; (track < file_header.tracks) && (buffer < buffer_end); ++track)
	{
		if (file_header.format != 2)
			current_absolute_time = 0;
			
		buffer = read_track_header(buffer);
		events = events->load_events(buffer, track_header.length, current_absolute_time);
	}

	events->set_absolute_times(file_header.division);

	return buffer;
}

unsigned char *libnotemidi_file::read_file_header(unsigned char *buffer)
{
	libnotemidi_character_file_header *data = (libnotemidi_character_file_header*) buffer;
	libnotemidi_event event;

	strncpy(file_header.chunk_type, data->chunk_type, sizeof(file_header.chunk_type));
	file_header.chunk_type[4] = (char)NULL;
	file_header.length = event.long_from_bytes(4, data->length);
	file_header.format = event.long_from_bytes(2, data->format);
	file_header.tracks = event.long_from_bytes(2, data->tracks);
	file_header.division = event.long_from_bytes(2, data->division);

	return (buffer + sizeof(libnotemidi_character_file_header));
}

unsigned char *libnotemidi_file::read_track_header(unsigned char *buffer)
{
	libnotemidi_character_track_header *data = (libnotemidi_character_track_header*) buffer;
	libnotemidi_event event;

	strncpy(track_header.chunk_type, data->chunk_type, sizeof(track_header.chunk_type));
	track_header.chunk_type[4] = (char)NULL;
	track_header.length = event.long_from_bytes(4, data->length);

	return (buffer + sizeof(libnotemidi_character_track_header));
}

unsigned char *libnotemidi_file::write_file_header(unsigned char *buffer)
{
	libnotemidi_character_file_header *data = (libnotemidi_character_file_header*)buffer;
	libnotemidi_event event;

	memset(data, (char)NULL, sizeof(libnotemidi_character_file_header));
	strncpy(data->chunk_type, file_header.chunk_type, sizeof(data->chunk_type));
	event.long_to_bytes(4, data->length, file_header.length);
	event.long_to_bytes(2, data->format, file_header.format);
	event.long_to_bytes(2, data->tracks, file_header.tracks);
	event.long_to_bytes(2, data->division, file_header.division);

	return (buffer + sizeof(libnotemidi_character_file_header));
}

unsigned char *libnotemidi_file::write_track_header(unsigned char *buffer, int track_length)
{
	libnotemidi_character_track_header *data = (libnotemidi_character_track_header*)buffer;
	libnotemidi_event event;

	track_header.length = track_length;
	strncpy(data->chunk_type, track_header.chunk_type, sizeof(data->chunk_type));
	event.long_to_bytes(4, data->length, track_header.length);

	return (buffer + sizeof(libnotemidi_character_track_header));
}

unsigned char *libnotemidi_file::write_end_of_track(unsigned char *buffer)
{
	libnotemidi_event event;
	event.delta_time = 0;
	event.data = new unsigned char[3];

	event.data[0] = LIBNOTEMIDI_META_EVENT;
	event.data[1] = LIBNOTEMIDI_META_END_OF_TRACK;
	event.data[2] = 0;

	return event.write_event(buffer);
}
